/* 
 * contr_bright_pgm.c 
 * version 1.1
 * author Felix Burkhardt
 * last change:  Sun Sep 20 17:52:04 GMT 1998   
 *
 * Routinen fue Kontrast- und Brightness-Veraenderungen.
 * fuer Greymaps
 *
 * Die Routinen nehem einen Pointer auf Beschreibungstukturen
 * fuer die Anderungen
 * Jede Struktur besteht aus Reihenangabe und Veraenderungsfaktor
 * Die Strukturen MUESSEN nach aufsteigenden Reihen sortiert sein
 * Der Faktor ist um 0.5 skaliert (zwischen 0 and 1).
 *
 * Um den Kontrast zu veraendern wird der Mittelwert des Pixel-Feldes
 * berechnet und als Kontrastgrenze verwendet.
 */

#include "contr_bright_pgm.h"
#include <stdio.h>

/* check overflow                                                          */
#define CHECK_MAX(x) (x>MAXVAL ? MAXVAL : x)
#define CHECK_MIN(x) (x<0 ? 0 : x)
#define CHECK_BORDER_MAX(x, b) (x>b ? b  : x)
#define CHECK_BORDER_MIN(x, b) (x<b ? b : x)

int changeContrastRows(pixval * pix_array,	/* inputChannel */
		       int minrow, int maxrow,	/* min und Maxrow  */
		       int width,
		       int border,     /* decisionborder for contrast: higher 
				          or lower? */
		       double start_val,	/* start-val contrast-change */
		       double end_val);		/* end-val contrast-change */

/* Zur Beschleunigung, falls keine Interpolation noetig */
int changeContrastRows_simple(pixval * pix_array,	/* inputChannel */
			      int minrow, int maxrow,	/* min und Maxrow  */
			      int width,
			      int border,	/* decisionborder for
						   contrast: higher or lower? 
						 */
			      double val);	/* val contrast-change */

int changeBrightnessRows(pixval * pix_array,	/* inputChannel */
			 int minrow, int maxrow,	/* min und Maxrow */
			 int width,
			 double start_val,	/* start-val
						   brightness-change */
			 double end_val);	/* end-val brightness-change */
/* Zur Beschleunigung, falls keine Interpolation noetig */
int changeBrightnessRows_simple(pixval * pix_array,	/* inputChannel */
				int minrow, int maxrow,		/* min und
								   Maxrow */
				int width,
				double val);	/* factor for
						   brightness-change */

extern pixval *changeContrast_PGM(pixval * pix_array_in,	/* inputPixmap 
								 */
				  int width, int height,
				  int changeDescrNum,	/* number of
							   change-descriptors 
							 */
				  changeDescriptorP cDPointer)
{				       /* change-descriptors */
    pixval
	* pix_array_out;
    int
      len, i, ok, border,	       /* decisionborder for contrast: higher 
				          or lower? */
      minrow, maxrow;
    double
      start_val, end_val;

    len = width * height;
    pix_array_out = (pixval *) malloc(len * sizeof(pixval));
    memcpy(pix_array_out, pix_array_in, len);
    border = (int) getMeanChannel(pix_array_in, len);

    for (i = 0; i < changeDescrNum; i++) {
	/* if there's only one field */
	if (changeDescrNum == 1) {
	    minrow = 0;
	    maxrow = height;
	    start_val = cDPointer[0].val;
	    if (start_val == 0.5)      /* no change, return */
		return pix_array_out;
	    end_val = start_val;
	}
	/* if it's the last field and ends in the picture change the last
	   rows */
	else if (i == changeDescrNum - 1 && cDPointer[i].row < height) {
	    minrow = cDPointer[i].row;
	    maxrow = height;
	    start_val = cDPointer[i].val;
	    end_val = start_val;
	}
	/* if it's the last field there's nothing to be done */
	else if (i == changeDescrNum - 1 && cDPointer[i].row >= height) {
	    break;
	}
	/* if we're in the field */
	else {
	    /* if it's the first field and starts in the picture change the
	       first rows */
	    if (i == 0 && cDPointer[0].row != 0) {
		minrow = 0;
		maxrow = cDPointer[0].row;
		start_val = cDPointer[i].val;
		if (start_val != 0.5) {		/* otherwise no change */
		    ok = changeContrastRows_simple(pix_array_out,
						   minrow, maxrow,
						   width,
						   border, start_val);
		}
	    }
	    minrow = cDPointer[i].row;
	    if ((maxrow = cDPointer[i + 1].row) > height)
		maxrow = height;
	    start_val = cDPointer[i].val;
	    end_val = cDPointer[i + 1].val;
	}
	if (end_val != start_val)
	    ok = changeContrastRows(pix_array_out, minrow, maxrow,
				    width,
				    border, start_val, end_val);
	else if (start_val != 0.5)
	    ok = changeContrastRows_simple(pix_array_out, minrow, maxrow,
					   width,
					   border, start_val);
    }

    return pix_array_out;
}

extern pixval *changeBrightness_PGM(pixval * pix_array_in,	/* inputPPM */
				    int width, int height,
				    int changeDescrNum,		/* number of
								   change-descriptors 
								 */
				    changeDescriptorP cDPointer)
{				       /* change-descriptors */
    pixval
	* pix_array_out;
    int
      len, i, ok, minrow, maxrow;
    double
      start_val, end_val;

    len = width * height;

    pix_array_out = (pixval *) malloc(len * sizeof(pixval));
    memcpy(pix_array_out, pix_array_in, len);

    for (i = 0; i < changeDescrNum; i++) {
	/* if there's only one field */
	if (changeDescrNum == 1) {
	    minrow = 0;
	    maxrow = height;
	    start_val = cDPointer[0].val;
	    if (start_val == 0.5)      /* no change, return */
		return pix_array_out;
	    end_val = start_val;
	}
	/* if it's the last field and ends in the picture change the last
	   rows */
	else if (i == changeDescrNum - 1 && cDPointer[i].row < height) {
	    minrow = cDPointer[i].row;
	    maxrow = height;
	    start_val = cDPointer[i].val;
	    end_val = start_val;
	}
	/* if it's the last field there's nothing to be done */
	else if (i == changeDescrNum - 1 && cDPointer[i].row >= height) {
	    break;
	}
	/* if we're in the field */
	else {
	    /* if it's the first field and starts in the picture change the
	       first rows */
	    if (i == 0 && cDPointer[0].row != 0) {
		minrow = 0;
		maxrow = cDPointer[0].row;
		start_val = cDPointer[i].val;
		if (start_val != 0.5) {		/* otherwise no change */
		    ok = changeBrightnessRows_simple(pix_array_out,
						     minrow, maxrow,
						     width,
						     start_val);
		}
	    }
	    minrow = cDPointer[i].row;
	    if ((maxrow = cDPointer[i + 1].row) > height)
		maxrow = height;
	    start_val = cDPointer[i].val;
	    end_val = cDPointer[i + 1].val;
	}
	if (end_val != start_val)
	    ok = changeBrightnessRows(
					 pix_array_out,
					 minrow, maxrow,
					 width,
					 start_val, end_val);
	else if (start_val != 0.5)
	    ok = changeBrightnessRows_simple(
						pix_array_out,
						minrow, maxrow,
						width,
						start_val);
    }
    return pix_array_out;
}

int changeContrastRows(pixval * pix_array,	/* pixChannel */
		       int minrow, int maxrow,	/* min und Maxrow */
		       int width,
		       int border,
		       double start_val,	/* start-val contrast-change */
		       double end_val)
{				       /* end-val contrast-change */
    int
      change, row, col, rowNum, actrow;
    double
      start_fac, end_fac, step_fac, act_fac;

    rowNum = maxrow - minrow;
    start_fac = (start_val - 0.5) * 2; /* scale factor around 0 */
    end_fac = (end_val - 0.5) * 2;     /* scale factor around 0         */

    step_fac = (double) (end_fac - start_fac) / rowNum;
    act_fac = start_fac;

    for (row = minrow; row < maxrow; row++) {
	/* actualize change */
	change = (MAXVAL - border) * act_fac;
	actrow = row * width;
	for (col = 0; col < width; col++) {
	    if (change > 0)	       /* enhance contrast */
		if (pix_array[(actrow + col)] >= border)
		    pix_array[(actrow + col)] = CHECK_MAX(pix_array[(actrow + col)] + change);
		else
		    pix_array[(actrow + col)] = CHECK_MIN(pix_array[(actrow + col)] - change);
	    else
		/* reduce contrast */
	    if (pix_array[(actrow + col)] >= border)
		pix_array[(actrow + col)] = CHECK_BORDER_MIN(pix_array[(actrow + col)] + change, border);
	    else
		pix_array[(actrow + col)] = CHECK_BORDER_MAX(pix_array[(actrow + col)] - change, border);
	}
	act_fac += step_fac;
    }
    return 0;
}

int changeContrastRows_simple(pixval * pix_array,	/* pixChannel */
			      int minrow, int maxrow,	/* min und Maxrow */
			      int width,
			      int border,
			      double val)
{				       /* val contrast-change */
    int
      change, act;
    double
      fac;

    fac = (val - 0.5) * 2;	       /* scale factor around 0 */
    change = (MAXVAL - border) * fac;

    for (act = minrow * width; act < maxrow * width; act++) {
	if (change > 0)		       /* enhance contrast */
	    if (pix_array[act] >= border)
		pix_array[act] = CHECK_MAX(pix_array[act] + change);
	    else
		pix_array[act] = CHECK_MIN(pix_array[act] - change);
	else
	    /* reduce contrast */
	if (pix_array[act] >= border)
	    pix_array[act] = CHECK_BORDER_MIN(pix_array[act] + change, border);
	else
	    pix_array[act] = CHECK_BORDER_MAX(pix_array[act] - change, border);
    }
    return 0;
}

int changeBrightnessRows(
			    pixval * pix_array,		/* pixChannel */
			    int minrow, int maxrow,	/* min und Maxrow */
			    int width,
			    double start_val,	/* factor for contrast-change 
						 */
			    double end_val)
{				       /* factor for contrast-change */
    int
      row, col, actrow, bright_change, rowNum;
    double
      start_fac, end_fac, step_fac, act_fac;

    rowNum = maxrow - minrow;
    start_fac = (start_val - 0.5) * 2; /* scale factor around 0 */
    end_fac = (end_val - 0.5) * 2;     /* scale factor around 0 */

    step_fac = (double) (end_fac - start_fac) / rowNum;
    act_fac = start_fac;

    for (row = minrow; row < maxrow; row++) {
	bright_change = MAXVAL * act_fac;
	actrow = row * width;
	for (col = 0; col < width; col++) {
	    if (bright_change > 0)     /* enhance brightness */
		pix_array[(actrow + col)] = CHECK_MAX(pix_array[(actrow + col)] + bright_change);
	    else		       /* reduce brightness               */
		pix_array[(actrow + col)] = CHECK_MIN(pix_array[(actrow + col)] + bright_change);
	}
	act_fac += step_fac;
    }
    return 0;
}
int changeBrightnessRows_simple(
				   pixval * pix_array,	/* pixChannel */
				   int minrow, int maxrow,	/* min und
								   Maxrow */
				   int width,
				   double val)
{				       /* factor for brightness-change */
    int
      act, bright_change;
    double
      fac;

    fac = (val - 0.5) * 2;	       /* scale factor around 0 */
    bright_change = MAXVAL * fac;
    for (act = minrow * width; act < maxrow * width; act++) {
	if (bright_change > 0)	       /* enhance brightness */
	    pix_array[act] = CHECK_MAX(pix_array[act] + bright_change);
	else			       /* reduce brightness               */
	    pix_array[act] = CHECK_MIN(pix_array[act] + bright_change);
    }
    return 0;
}
